#################################################################################################
## type mapping...
#################################################################################################
my $gen_types = {
    'bool' => {
        'str' => 0,
        'sql' => 'BIT',
        'cs' => 'Boolean',
        'csdb' => 'SqlDbType.Bit',
        'dispsize' => 32,
    },
    'uniqueidentifier' => {
        'str' => 0,
        'sql' => 'uniqueidentifier',
        'cs' => 'Guid',
        'dispsize' => 128,
    },
    'int' => {
        'str' => 0,
        'sql' => 'INT',
        'cs' => 'Int32',
        'csdb' => 'SqlDbType.Int',
        'dispsize' => 64,
    },
    'uint' => {
        'str' => 0,
        'sql' => 'INT UNSIGNED',
        'dispsize' => 64,
    },
    'char' => {
        'str' => 1,
        'sql' => 'CHAR',
        'cs' => 'String',
        'csdb' => 'SqlDbType.Char',
        'dispsize' => 5,
    },
    'string' => {
        'str' => 1,
        'sql' => 'VARCHAR',
        'cs' => 'String',
        'csdb' => 'SqlDbType.VarChar',
        'dispsize' => 5,
    },
    'text' => {
        'str' => 1,
        'sql' => 'TEXT',
        'cs' => 'String',
        'csdb' => 'SqlDbType.Text',
        'dispsize' => 5,
    },
    'date' => {
        'str' => 1,
        'sql' => 'DATE',
        'dispsize' => 128,
    },
    'time' => {
        'str' => 1,
        'sql' => 'TIME',
        'dispsize' => 128,
    },
    'datetime' => {
        'str' => 1,
        'sql' => 'DATETIME',
        'cs' => 'DateTime',
        'csdb' => 'SqlDbType.DateTime',
        'dispsize' => 128,
    },
    'money' => {
        'str' => 0,
        'sql' => 'MONEY',
        'cs' => 'Decimal',
        'csdb' => 'SqlDbType.Money',
        'dispsize' => 64,
    },
};



#################################################################################################
## load file, parse xml, etc.
#################################################################################################

my $root = parseXML(loadFile("metadata.xml"));

#################################################################################################
## turn XML into meta objects
#################################################################################################

my @objects;
for $object (queryXML($root,'/meta/@object')){
    push @objects,make_object($object);
    $objects{ $objects[-1]->{name} } = $objects[-1];
}


#################################################################################################
## generate code
#################################################################################################

make_sql(\@objects);

make_value_objects(\@objects);

make_db_proxy_objects(\@objects);

make_listview_form(\@objects);

make_detailview_form(\@objects);

make_choosermenuview_form(\@objects);


#################################################################################################
## make_choosermenuview_form
#################################################################################################
sub make_choosermenuview_form {
   my ($objects) = @_;

   open my $out,'>gen/RawChooserMenuView.cs' or die $!;

print $out qq|
/**
 * RawChooserMenuView.cs
 *
 * THIS FILE IS AUTO-GENERATED FROM metadata.xml.
 * DO NOT UPDATE THIS FILE.
 * ANY CHANGES TO THIS FILE WILL BE LOST!!!
 * ALL CHANGES MUST BE MADE TO metadata.xml.
 */
using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;

namespace SAMPLEAPP
{
    /**
     * RawChooserMenuView, chooser menu for all the generated forms.
     */
    public class RawChooserMenuView : System.Windows.Forms.Form
    {
|;
   foreach $object (@{$objects}){
print $out qq|
        private System.Windows.Forms.Panel $object->{name}Panel;
        private System.Windows.Forms.Button manage$object->{name}Button;
|;
      }
print $out qq|

        private System.Windows.Forms.Panel buttonPanel;
        private System.Windows.Forms.Panel inputPanel;
        private System.Windows.Forms.Button closeButton;

        // Required designer variable.
        private System.ComponentModel.Container components = null;

        public RawChooserMenuView()
        {
            // Required for Windows Form Designer support
            InitializeComponent();
        }

        /**
         * Clean up any resources being used.
         */
        protected override void Dispose( bool disposing )
        {
            if( disposing ) {
                if(components != null) {
                    components.Dispose();
                }
            }
            base.Dispose( disposing );
        }

        #region Windows Form Designer generated code

        /**
         * Required method for Designer support - do not modify
         * the contents of this method with the code editor.
         */
        private void InitializeComponent()
        {
|;
   foreach $object (@{$objects}){
print $out qq|
            this.$object->{name}Panel = new System.Windows.Forms.Panel();
            this.manage$object->{name}Button = new System.Windows.Forms.Button();
            this.$object->{name}Panel.SuspendLayout();
|;
      }
print $out qq|
            this.closeButton = new System.Windows.Forms.Button();

            this.buttonPanel = new System.Windows.Forms.Panel();
            this.inputPanel = new System.Windows.Forms.Panel();

            this.buttonPanel.SuspendLayout();
            this.inputPanel.SuspendLayout();

            this.SuspendLayout();
|;
   foreach $object (@{$objects}){
print $out qq|
            this.$object->{name}Panel = new System.Windows.Forms.Panel();
            this.manage$object->{name}Button = new System.Windows.Forms.Button();
            this.$object->{name}Panel.SuspendLayout();

            // manage$object->{name}Button
            this.manage$object->{name}Button.Dock = System.Windows.Forms.DockStyle.Left;
            this.manage$object->{name}Button.Location = new System.Drawing.Point(152, 10);
            this.manage$object->{name}Button.Name = "manage$object->{name}Button";
            this.manage$object->{name}Button.Size = new System.Drawing.Size(208, 20);
            this.manage$object->{name}Button.TabIndex = 3;
            this.manage$object->{name}Button.Text = "Manage $object->{pname}";
            this.manage$object->{name}Button.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.manage$object->{name}Button.Click += new System.EventHandler(this.manage$object->{name}Button_Click);

            // $object->{name}Panel
            this.$object->{name}Panel.Controls.AddRange(
                new System.Windows.Forms.Control[] {
                    this.manage$object->{name}Button
                }
            );

            this.$object->{name}Panel.Dock = System.Windows.Forms.DockStyle.Top;
            this.$object->{name}Panel.DockPadding.All = 10;
            this.$object->{name}Panel.Name = "$object->{name}Panel";
            this.$object->{name}Panel.Size = new System.Drawing.Size(562, 40);
            this.$object->{name}Panel.TabIndex = 6;

|;
      }
print $out qq|

            // cancelButton
            this.closeButton.Anchor = (System.Windows.Forms.AnchorStyles.Top \| System.Windows.Forms.AnchorStyles.Right);
            this.closeButton.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.closeButton.Location = new System.Drawing.Point(376, 24);
            this.closeButton.Name = "closeButton";
            this.closeButton.TabIndex = 5;
            this.closeButton.Text = "Close";
            this.closeButton.Click += new System.EventHandler(this.closeButton_Click);

            // buttonPanel
            this.buttonPanel.Controls.AddRange(
                new System.Windows.Forms.Control[] {
                    this.closeButton,
                }
            );

            this.buttonPanel.Dock = System.Windows.Forms.DockStyle.Bottom;
            this.buttonPanel.Location = new System.Drawing.Point(0, 280);
            this.buttonPanel.Name = "buttonPanel";
            this.buttonPanel.Size = new System.Drawing.Size(562, 64);
            this.buttonPanel.TabIndex = 8;

            // inputPanel
            this.inputPanel.AutoScroll = true;
            this.inputPanel.AutoScrollMargin = new System.Drawing.Size(5, 5);
            this.inputPanel.AutoScrollMinSize = new System.Drawing.Size(300, 200);
            this.inputPanel.DockPadding.All = 8;

            this.inputPanel.Controls.AddRange(
                new System.Windows.Forms.Control[] {
|;
               my @arr = ();
               foreach $object (reverse @{$objects}){
                  push @arr,"this.$object->{name}Panel";
               }
               print $out join(",\n                    ",@arr);
print $out qq|
                }
            );

            this.inputPanel.Dock = System.Windows.Forms.DockStyle.Fill;
            this.inputPanel.Name = "inputPanel";
            this.inputPanel.Size = new System.Drawing.Size(562, 280);
            this.inputPanel.TabIndex = 9;

            // Raw$object->{name}AddView
            this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
            this.ClientSize = new System.Drawing.Size(562, 344);
            this.Controls.AddRange(
                new System.Windows.Forms.Control[] {
                    this.inputPanel,
                    this.buttonPanel
                }
            );

            this.Name = "RawChooserMenuForm";
            this.StartPosition = System.Windows.Forms.FormStartPosition.CenterParent;
            this.Text = "Choose What You Want To Do!";
|;
   foreach $object (@{$objects}){
print $out qq|
            this.$object->{name}Panel.ResumeLayout(false);
|;
      }
print $out qq|
            this.buttonPanel.ResumeLayout(false);
            this.inputPanel.ResumeLayout(false);
            this.ResumeLayout(false);

            this.inputPanel.PerformLayout();

        }
        #endregion

        private void closeButton_Click(object sender, System.EventArgs e)
        {
            this.Close();
        }
|;
      foreach $object (@{$objects}){
print $out qq|
        private void manage$object->{name}Button_Click(object sender, System.EventArgs e)
        {
            new Raw$object->{name}ListView().Show();
        }
|;
      }
print $out qq|
    }
}
|;
      close $out;
}









#################################################################################################
## make_detailview_form
#################################################################################################
sub make_detailview_form {
   my ($objects) = @_;

   foreach $object (@{$objects}){

      open my $out,'>gen/Raw'.$object->{name}.'DetailView.cs' or die $!;


      foreach $attrib (@{$object->{attrib}}){
         if($attrib->{oneof}){
            $attrib->{attribute_name} = "$attrib->{name}ComboBox";
            $attrib->{attribute_type} = "System.Windows.Forms.ComboBox";
         }else{
            $attrib->{attribute_name} = "$attrib->{name}TextBox";
            $attrib->{attribute_type} = "System.Windows.Forms.TextBox";
         }
      }


print $out qq|
/**
 * Raw$object->{name}DetailView.cs
 *
 * THIS FILE IS AUTO-GENERATED FROM metadata.xml.
 * DO NOT UPDATE THIS FILE.
 * ANY CHANGES TO THIS FILE WILL BE LOST!!!
 * ALL CHANGES MUST BE MADE TO metadata.xml.
 */
using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;

namespace SAMPLEAPP
{
    /**
     * $object->{name} Detail View object; allows users to create/update.
     */
    public class Raw$object->{name}DetailView : System.Windows.Forms.Form
    {

        private Raw$object->{name}ListView theParent;
        private bool updateView;

|;
      foreach $attrib (@{$object->{attrib}}){
print $out qq|
        private System.Windows.Forms.Panel $attrib->{name}Panel;
        private System.Windows.Forms.Label $attrib->{name}Label;
        private $attrib->{attribute_type} $attrib->{attribute_name};
|;
      }

print $out qq|

        private System.Windows.Forms.Panel buttonPanel;
        private System.Windows.Forms.Panel inputPanel;
        private System.Windows.Forms.Button cancelButton;
        private System.Windows.Forms.Button actionButton;

        // Required designer variable.
        private System.ComponentModel.Container components = null;

        /**
         * hepler method to setup stuff
         */
        private void SetupNamesColorsToolTips(){
            if(updateView){
                this.actionButton.Text = "Update";
                this.Text = "Update $object->{fname}";
            } else {
                this.actionButton.Text = "Create";
                this.Text = "Create New $object->{fname}";
            }

            // indicate required fields.
|;
      foreach $attrib (@{$object->{attrib}}){
         if($attrib->{required}){
   print $out qq|
            $attrib->{attribute_name}.BackColor = System.Drawing.Color.Ivory;
   |;
         }
      }
print $out qq|
            // Create the ToolTip and associate with the Form container.
            ToolTip toolTip1 = new ToolTip();

            // Set up the delays for the ToolTip.
            toolTip1.AutoPopDelay = 5000;
            toolTip1.InitialDelay = 1000;
            toolTip1.ReshowDelay = 500;

            // Force the ToolTip text to be displayed whether or not the form is active.
            toolTip1.ShowAlways = true;

|;
      foreach $attrib (@{$object->{attrib}}){
            my $required = "(Required)" if $attrib->{required};
print $out qq|
            toolTip1.SetToolTip(this.$attrib->{name}Label, "$attrib->{description} $required");
            toolTip1.SetToolTip(this.$attrib->{attribute_name}, "$attrib->{description} $required");
|;
      }
print $out qq|
            toolTip1.SetToolTip(this.cancelButton, "Close this form.");
            toolTip1.SetToolTip(this.actionButton, "Perform the operation.");
|;

      foreach $attrib (@{$object->{attrib}}){
         if($attrib->{oneof}){
            my $robj = $objects{ $attrib->{oneof} };
print $out qq|
            $attrib->{attribute_name}.BeginUpdate();
            $attrib->{attribute_name}.DisplayMember = "$robj->{display}{cspname}";
            $attrib->{attribute_name}.ValueMember = "$robj->{value}{cspname}";

            Int32 default$attrib->{attribute_name} = 0;
            if(Global.prefs.Contains("Default$robj->{name}")){
                default$attrib->{attribute_name} = Convert.ToInt32( Global.prefs["Default$robj->{name}"] );
            }
|;
          if(!$attrib->{required}){
print $out qq|
            $robj->{name} blank$attrib->{name} = new $robj->{name}();
            blank$attrib->{name}.$robj->{display}{cspname} = "";
            blank$attrib->{name}.$robj->{value}{cspname} = 0;
            $attrib->{attribute_name}.Items.Add(blank$attrib->{name});
|;
          }

print $out qq|
            for(int i=0;i<Global.$robj->{globalref}.Count;i++){
                $robj->{name} o = ($robj->{name})Global.$robj->{globalref} [i];
                $attrib->{attribute_name}.Items.Add(o);
                if(o.$robj->{value}{cspname} == default$attrib->{attribute_name}){
                    $attrib->{attribute_name}.SelectedIndex = $attrib->{attribute_name}.Items.Count - 1;
                }
            }
            if($attrib->{attribute_name}.SelectedIndex < 0)
                $attrib->{attribute_name}.SelectedIndex = 0;
            $attrib->{attribute_name}.EndUpdate();
|;
         }
      }

print $out qq|
        }
|;

print $out qq|
        /**
         * Used for the update of $object->{name} object; id indicates primary key.
         */
        public Raw$object->{name}DetailView(Raw$object->{name}ListView p, ${$object->{attrib_names}}{$object->{firstkey}}->{cstype} id)
        {
            theParent = p;
            updateView = true;
            // Required for Windows Form Designer support
            InitializeComponent();
            SetupNamesColorsToolTips();
            LoadData(id);
        }

        /**
         * Used for the creation of a $object->{name} object.
         */
        public Raw$object->{name}DetailView(Raw$object->{name}ListView p)
        {
            theParent = p;
            updateView = false;
            // Required for Windows Form Designer support
            InitializeComponent();
            SetupNamesColorsToolTips();
        }

        /**
         * allows an update form to load the data from the database via
         * a primary key.
         */
        private void LoadData(${$object->{attrib_names}}{$object->{firstkey}}->{cstype} id)
        {
            try{
               $object->{name} o = $object->{name}DB.Get$object->{name}(id);
|;

      foreach $attrib (@{$object->{attrib}}){

         if($attrib->{oneof}){

            my $robj = $objects{ $attrib->{oneof} };
print $out qq|
            $attrib->{attribute_name}.BeginUpdate();
            for(int i=0;i<$attrib->{attribute_name}.Items.Count;i++){
                if((($robj->{name})$attrib->{attribute_name}.Items[i]).$robj->{value}{cspname} == o.$attrib->{cspname}){
                    $attrib->{attribute_name}.SelectedIndex = i;
                    break;
                }
            }
            $attrib->{attribute_name}.EndUpdate();
|;
         }else{
print $out qq|
               $attrib->{name}TextBox.Text = Convert.ToString( o.$attrib->{cspname} );
|;
         }

      }
print $out qq|

            }catch(ApplicationException ex){
                MessageBox.Show(ex.Message, "Application Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                this.Close();
            }catch(Exception ex){
                MessageBox.Show(ex.Message, "Serious Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                this.Close();
            }
        }

        /**
         * Clean up any resources being used.
         */
        protected override void Dispose( bool disposing )
        {
            if( disposing ){
                if(components != null){
                    components.Dispose();
                }
            }
            base.Dispose( disposing );
        }

        #region Windows Form Designer generated code
        /**
         * Required method for Designer support - do not modify
         * the contents of this method with the code editor.
         */
        private void InitializeComponent()
        {
|;

      foreach $attrib (@{$object->{attrib}}){
print $out qq|
            this.$attrib->{name}Panel = new System.Windows.Forms.Panel();
            this.$attrib->{name}Label = new System.Windows.Forms.Label();
            this.$attrib->{attribute_name} = new $attrib->{attribute_type}();
            this.$attrib->{name}Panel.SuspendLayout();
|;
      }

print $out qq|

            this.actionButton = new System.Windows.Forms.Button();
            this.cancelButton = new System.Windows.Forms.Button();

            this.buttonPanel = new System.Windows.Forms.Panel();
            this.inputPanel = new System.Windows.Forms.Panel();

            this.buttonPanel.SuspendLayout();
            this.inputPanel.SuspendLayout();

            this.SuspendLayout();
|;

      my $i = 0;
      foreach $attrib (@{$object->{attrib}}){
         if($attrib->{oneof}){
            $attrib->{dispsize} = $objects{$attrib->{oneof}}{display}{dispsize};
         }

print $out qq|

            // $attrib->{name}Label
            this.$attrib->{name}Label.AutoSize = true;
            this.$attrib->{name}Label.Dock = System.Windows.Forms.DockStyle.Left;
            this.$attrib->{name}Label.Location = new System.Drawing.Point(10, 10);
            this.$attrib->{name}Label.Name = "$attrib->{name}Label";
            this.$attrib->{name}Label.Size = new System.Drawing.Size(142, 13);
            this.$attrib->{name}Label.Text = "$attrib->{fname}:";
            this.$attrib->{name}Label.TextAlign = System.Drawing.ContentAlignment.MiddleRight;

            // $attrib->{attribute_name}
            this.$attrib->{attribute_name}.Dock = System.Windows.Forms.DockStyle.Left;
            this.$attrib->{attribute_name}.Location = new System.Drawing.Point(152, 10);
            this.$attrib->{attribute_name}.Name = "$attrib->{attribute_name}";
            this.$attrib->{attribute_name}.Size = new System.Drawing.Size($attrib->{dispsize}, 20);
            this.$attrib->{attribute_name}.TabIndex = 0;
            this.$attrib->{attribute_name}.Text = "";

            // $attrib->{name}Panel
            this.$attrib->{name}Panel.Controls.AddRange(
                new System.Windows.Forms.Control[] {
                    this.$attrib->{attribute_name},
                    this.$attrib->{name}Label
                }
            );
            this.$attrib->{name}Panel.Dock = System.Windows.Forms.DockStyle.Top;
            this.$attrib->{name}Panel.DockPadding.All = 10;
            this.$attrib->{name}Panel.Name = "$attrib->{name}Panel";
            this.$attrib->{name}Panel.Size = new System.Drawing.Size(562, 40);
            this.$attrib->{name}Panel.TabIndex = $i;
|;
         if($attrib->{auto}){
print $out qq|
            this.$attrib->{attribute_name}.Enabled = false;
|;
         }

         if($attrib->{oneof}){
print $out qq|
            this.$attrib->{attribute_name}.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
|;
         }


         $i++;
      }

print $out qq|

            // actionButton
            this.actionButton.Anchor = (System.Windows.Forms.AnchorStyles.Top \| System.Windows.Forms.AnchorStyles.Right);
            this.actionButton.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.actionButton.Location = new System.Drawing.Point(472, 24);
            this.actionButton.Name = "updateButton";
            this.actionButton.TabIndex = 2;

            this.actionButton.Text = "Do Not Change!";

            this.actionButton.Click += new System.EventHandler(this.actionButton_Click);

            // cancelButton
            this.cancelButton.Anchor = (System.Windows.Forms.AnchorStyles.Top \| System.Windows.Forms.AnchorStyles.Right);
            this.cancelButton.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.cancelButton.Location = new System.Drawing.Point(376, 24);
            this.cancelButton.Name = "cancelButton";
            this.cancelButton.TabIndex = 1;
            this.cancelButton.Text = "Cancel";
            this.cancelButton.Click += new System.EventHandler(this.cancelButton_Click);

            // buttonPanel
            this.buttonPanel.Controls.AddRange(
                new System.Windows.Forms.Control[] {
                    this.cancelButton,
                    this.actionButton
                }
            );

            this.buttonPanel.Dock = System.Windows.Forms.DockStyle.Bottom;
            this.buttonPanel.Location = new System.Drawing.Point(0, 280);
            this.buttonPanel.Name = "buttonPanel";
            this.buttonPanel.Size = new System.Drawing.Size(562, 64);
            this.buttonPanel.TabIndex = 1;

            // inputPanel
            this.inputPanel.AutoScroll = true;
            this.inputPanel.AutoScrollMargin = new System.Drawing.Size(5, 5);
            this.inputPanel.AutoScrollMinSize = new System.Drawing.Size(300, 200);
            this.inputPanel.DockPadding.All = 16;


            this.inputPanel.Controls.AddRange(
                new System.Windows.Forms.Control[] {
|;
               my @arr = ();
               foreach $attrib (reverse @{$object->{attrib}}){
                  push @arr,"this.$attrib->{name}Panel";
               }
               print $out join(",\n                    ",@arr);
print $out qq|

                }
            );

            this.inputPanel.Dock = System.Windows.Forms.DockStyle.Fill;
            this.inputPanel.Name = "inputPanel";
            this.inputPanel.Size = new System.Drawing.Size(562, 280);
            this.inputPanel.TabIndex = 2;

            // Raw$object->{name}AddView
            this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
            this.ClientSize = new System.Drawing.Size(562, 344);
            this.Controls.AddRange(
                new System.Windows.Forms.Control[] {
                    this.inputPanel,
                    this.buttonPanel
                }
            );

            this.Name = "Raw$object->{name}DetailView";
            this.StartPosition = System.Windows.Forms.FormStartPosition.CenterParent;

|;
      foreach $attrib (@{$object->{attrib}}){
print $out qq|
            this.$attrib->{name}Panel.ResumeLayout(false);
|;
      }
print $out qq|

            this.buttonPanel.ResumeLayout(false);
            this.inputPanel.ResumeLayout(false);
            this.ResumeLayout(false);

            this.inputPanel.PerformLayout();

        }
        #endregion

        private void cancelButton_Click(object sender, System.EventArgs e)
        {
            this.Close();
        }

        private void actionButton_Click(object sender, System.EventArgs e)
        {
            // call the server and reload parent window.
            $object->{name} o = new $object->{name}();
|;

      foreach $attrib (@{$object->{attrib}}){

          if($attrib->{oneof}){
print $out qq|
            if($attrib->{attribute_name}.SelectedIndex >= 0)
            {
                o.$attrib->{cspname} = Convert.To$attrib->{cstype}( (( $objects{$attrib->{oneof}}->{name} )$attrib->{attribute_name}.Items[$attrib->{attribute_name}.SelectedIndex]).$objects{$attrib->{oneof}}->{value}{cspname} );
            }|;
          }else{
print $out qq|
            $attrib->{attribute_name}.Text = $attrib->{attribute_name}.Text.Trim();
            if($attrib->{attribute_name}.Text.Length > 0)
            {
                o.$attrib->{cspname} = Convert.To$attrib->{cstype}( $attrib->{attribute_name}.Text );
            }|;
          }

          if($attrib->{required}){
print $out qq|
            else
            {
                MessageBox.Show("The $attrib->{fname} field is required..." , "Application Error",MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }
|;
          }
      }
print $out qq|
            bool closeWindow = true;
            try{
                if(updateView){
                    $object->{name}DB.Update$object->{name}(o);
                } else {
                    $object->{name}DB.Create$object->{name}(o);
                }
                this.theParent.ReloadData();
            }catch(ApplicationException ex){
                closeWindow = false;
                MessageBox.Show(ex.Message, "Application Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }catch(Exception ex){
                MessageBox.Show(ex.Message, "Serious Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
            if(closeWindow){
                this.Close();
            }
        }
    }
}
|;
      close $out;
   }
}









#################################################################################################
## make_listview_form
#################################################################################################
sub make_listview_form {
   my ($objects) = @_;

   foreach $object (@{$objects}){

      open my $out,'>gen/Raw'.$object->{name}.'ListView.cs' or die $!;


print $out qq|
/*
 * Raw$object->{name}ListView.cs
 *
 * THIS FILE IS AUTO-GENERATED FROM metadata.xml.
 * DO NOT UPDATE THIS FILE.
 * ANY CHANGES TO THIS FILE WILL BE LOST!!!
 * ALL CHANGES MUST BE MADE TO metadata.xml.
 */

using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;

namespace SAMPLEAPP
{
    /**
     * ListView for $object->{name} objects.
     */
    public class Raw$object->{name}ListView : System.Windows.Forms.Form
    {
        private System.Windows.Forms.Panel panel1;
        private System.Windows.Forms.ListView listView1;
|;

      foreach $attrib (@{$object->{attrib}}){
print $out qq|
        private System.Windows.Forms.ColumnHeader $attrib->{name}ColumnHeader;
|;
      }

print $out qq|
        private System.Windows.Forms.Panel panel2;
        private System.Windows.Forms.Button closeButton;
        private System.Windows.Forms.Button removeButton;
        private System.Windows.Forms.Button updateButton;
        private System.Windows.Forms.Button addButton;

        // Required designer variable.
        private System.ComponentModel.Container components = null;

        public Raw$object->{name}ListView()
        {
            // Required for Windows Form Designer support
            InitializeComponent();

            // our method that loads the data into this component.
            LoadData();


            // Create the ToolTip and associate with the Form container.
            ToolTip toolTip1 = new ToolTip();

            // Set up the delays for the ToolTip.
            toolTip1.AutoPopDelay = 5000;
            toolTip1.InitialDelay = 1000;
            toolTip1.ReshowDelay = 500;

            // Force the ToolTip text to be displayed whether or not the form is active.
            toolTip1.ShowAlways = true;

            toolTip1.SetToolTip(this.closeButton, "Close this form.");
            toolTip1.SetToolTip(this.removeButton, "Remove selected $object->{fname} record.");
            toolTip1.SetToolTip(this.updateButton, "Update selected $object->{fname} record.");
            toolTip1.SetToolTip(this.addButton, "Add a new $object->{fname} record.");

        }

        /**
         * Clean up any resources being used.
         */
        protected override void Dispose( bool disposing )
        {
            if( disposing ) {
                if (components != null) {
                    components.Dispose();
                }
            }
            base.Dispose( disposing );
        }

        /**
         * method to load the data into this list.
         */
        private void LoadData(){
            ArrayList arr = $object->{name}DB.GetAll$object->{name}s();
            listView1.BeginUpdate();

            listView1.Items.Clear();

            foreach($object->{name} o in arr){
                System.Windows.Forms.ListViewItem listViewItem =
                    new System.Windows.Forms.ListViewItem(
                        new System.Windows.Forms.ListViewItem.ListViewSubItem[] {
|;
      my @arr = ();
      foreach $attrib (@{$object->{attrib}}){
         push @arr,qq|
                            new System.Windows.Forms.ListViewItem.ListViewSubItem(
                                null,
                                Convert.ToString(o.$attrib->{cspname})
                            )|;
      }
      print $out join(",\n",@arr);

print $out qq|

                        },
                        -1
                    );
                listViewItem.Tag = o.${$object->{attrib_names}}{$object->{firstkey}}->{cspname};
                listView1.Items.Add(listViewItem);
            }

            listView1.ListViewItemSorter = new ListViewItemComparer(0,listView1.Sorting);
            listView1.Sort();

            listView1.EndUpdate();
        }

        /**
         * reload data from the server, ie: 'refresh'
         */
        public void ReloadData()
        {
            // TODO: Start a new thread here (use invoke?)
            LoadData();
        }

        #region Windows Form Designer generated code

        /**
         * Required method for Designer support - do not modify
         * the contents of this method with the code editor.
         */
        private void InitializeComponent()
        {
            this.panel1 = new System.Windows.Forms.Panel();
            this.listView1 = new System.Windows.Forms.ListView();
|;

      foreach $attrib (@{$object->{attrib}}){
print $out qq|
            this.$attrib->{name}ColumnHeader = new System.Windows.Forms.ColumnHeader();
|;
      }
print $out qq|
            this.panel2 = new System.Windows.Forms.Panel();
            this.closeButton = new System.Windows.Forms.Button();
            this.removeButton = new System.Windows.Forms.Button();
            this.updateButton = new System.Windows.Forms.Button();
            this.addButton = new System.Windows.Forms.Button();
            this.panel1.SuspendLayout();
            this.panel2.SuspendLayout();
            this.SuspendLayout();

            // panel1
            this.panel1.Controls.AddRange(
                new System.Windows.Forms.Control[] {
                    this.listView1
                }
            );

            this.panel1.Dock = System.Windows.Forms.DockStyle.Fill;
            this.panel1.DockPadding.All = 16;
            this.panel1.Name = "panel1";
            this.panel1.Size = new System.Drawing.Size(392, 294);
            this.panel1.TabIndex = 0;

            // listView1
            this.listView1.AllowColumnReorder = true;
            this.listView1.Columns.AddRange(
                new System.Windows.Forms.ColumnHeader[] {
|;
      my @arr = ();
      foreach $attrib (@{$object->{attrib}}){
push @arr,qq|
                    this.$attrib->{name}ColumnHeader|;
      }
      print $out join(",",@arr);

print $out qq|
                }
            );

            this.listView1.ColumnClick += new System.Windows.Forms.ColumnClickEventHandler(this.listView1_ColumnClick);
            this.listView1.DoubleClick += new System.EventHandler(this.updateButton_Click);

            this.listView1.Sorting = System.Windows.Forms.SortOrder.Ascending;

            this.listView1.Dock = System.Windows.Forms.DockStyle.Fill;
            this.listView1.FullRowSelect = true;
            this.listView1.HideSelection = false;

            this.listView1.Location = new System.Drawing.Point(16, 16);
            this.listView1.MultiSelect = false;
            this.listView1.Name = "listView1";
            this.listView1.Size = new System.Drawing.Size(360, 262);
            this.listView1.TabIndex = 0;
            this.listView1.View = System.Windows.Forms.View.Details;
|;
      foreach $attrib (@{$object->{attrib}}){
print $out qq|
            this.$attrib->{name}ColumnHeader.Text = "$attrib->{fname}";
            this.$attrib->{name}ColumnHeader.Width = $attrib->{dispsize};
|;
      }
print $out qq|

            // panel2
            this.panel2.Controls.AddRange(
                new System.Windows.Forms.Control[] {
                    this.closeButton,
                    this.removeButton,
                    this.updateButton,
                    this.addButton
                }
            );

            this.panel2.Dock = System.Windows.Forms.DockStyle.Bottom;
            this.panel2.Location = new System.Drawing.Point(0, 294);
            this.panel2.Name = "panel2";
            this.panel2.Size = new System.Drawing.Size(392, 40);
            this.panel2.TabIndex = 1;

            // closeButton
            this.closeButton.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.closeButton.Location = new System.Drawing.Point(304, 0);
            this.closeButton.Name = "closeButton";
            this.closeButton.TabIndex = 0;
            this.closeButton.Text = "Close";
            this.closeButton.Click += new System.EventHandler(this.closeButton_Click);

            // removeButton
            this.removeButton.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.removeButton.Location = new System.Drawing.Point(208, 0);
            this.removeButton.Name = "removeButton";
            this.removeButton.TabIndex = 3;
            this.removeButton.Text = "Remove";
            this.removeButton.Click += new System.EventHandler(this.removeButton_Click);

            // updateButton
            this.updateButton.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.updateButton.Location = new System.Drawing.Point(112, 0);
            this.updateButton.Name = "updateButton";
            this.updateButton.TabIndex = 2;
            this.updateButton.Text = "Update";
            this.updateButton.Click += new System.EventHandler(this.updateButton_Click);

            // addButton
            this.addButton.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.addButton.Location = new System.Drawing.Point(16, 0);
            this.addButton.Name = "addButton";
            this.addButton.TabIndex = 1;
            this.addButton.Text = "Add";
            this.addButton.Click += new System.EventHandler(this.addButton_Click);

            // Form1
            this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
            this.ClientSize = new System.Drawing.Size(392, 334);

            this.Controls.AddRange(
                new System.Windows.Forms.Control[] {
                    this.panel1,
                    this.panel2
                }
            );

            this.MinimumSize = new System.Drawing.Size(400, 360);
            this.Name = "$object->{name}ListViewForm";
            this.StartPosition = System.Windows.Forms.FormStartPosition.CenterParent;
            this.Text = "Manage $object->{pname}";
            this.panel1.ResumeLayout(false);
            this.panel2.ResumeLayout(false);
            this.ResumeLayout(false);
        }
        #endregion

        /**
         * support for sorting by columns
         */
        private void listView1_ColumnClick(object sender, System.Windows.Forms.ColumnClickEventArgs e)
        {
            if (listView1.Sorting == SortOrder.None){
                listView1.Sorting = SortOrder.Ascending;
            } else if (listView1.Sorting == SortOrder.Ascending){
                listView1.Sorting = SortOrder.Descending;
            } else {
                listView1.Sorting = SortOrder.Ascending;
            }
            listView1.ListViewItemSorter = new ListViewItemComparer(e.Column, listView1.Sorting);
            listView1.Sort();
        }
|;

my $detailview = 'Raw'.$object->{name}.'DetailView';
if(-f 'Client/Raw'.$object->{name}.'DetailViewImpl.cs'){
   $detailview = 'Raw'.$object->{name}.'DetailViewImpl';
}

print $out qq|

        private void addButton_Click(object sender, System.EventArgs e)
        {
            new $detailview(this).ShowDialog(this);
        }

        private void updateButton_Click(object sender, System.EventArgs e)
        {
            foreach(ListViewItem lvi in listView1.SelectedItems){
                ${$object->{attrib_names}}{$object->{firstkey}}->{cstype} id = (${$object->{attrib_names}}{$object->{firstkey}}->{cstype})lvi.Tag;
                new $detailview(this,id).ShowDialog(this);
            }
        }

        private void removeButton_Click(object sender, System.EventArgs e)
        {
            foreach(ListViewItem lvi in listView1.SelectedItems){
                ${$object->{attrib_names}}{$object->{firstkey}}->{cstype} id = (${$object->{attrib_names}}{$object->{firstkey}}->{cstype})lvi.Tag;
                try{
                    if(DialogResult.Yes == MessageBox.Show(
                        "Are you sure you want to remove $object->{fname} record: "+Convert.ToString(id),
                        "Remove?",
                        MessageBoxButtons.YesNo,
                        MessageBoxIcon.Question))
                    {
                        $object->{name}DB.Delete$object->{name}(id);
                    }
                }catch(ApplicationException ex){
                    MessageBox.Show(
                        ex.Message,
                        "Application Error",
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Exclamation
                    );
                }catch(Exception ex){
                    MessageBox.Show(
                        ex.Message,
                        "Serious Error",
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Exclamation
                    );
                }
            }
            ReloadData();
        }

        private void closeButton_Click(object sender, System.EventArgs e)
        {
            this.Close();
        }

        class ListViewItemComparer : IComparer
        {
            private int col;
            private SortOrder sort;

            public ListViewItemComparer()
            {
                col=0;
                sort = SortOrder.None;
            }

            public ListViewItemComparer(int column)
            {
                col=column;
                sort = SortOrder.None;
            }

            public ListViewItemComparer(int column, SortOrder sortorder)
            {
                col=column;
                sort = sortorder;
            }

            public int Compare(object x, object y)
            {
                switch(col){
|;
      my %types;
      for(my $i=0;$i<=$#{$object->{attrib}};$i++){
          if(defined $types{ $object->{attrib}[$i]{cstype} }){
             push @{ $types{ $object->{attrib}[$i]{cstype} } },$i;
          }else{
             $types{ $object->{attrib}[$i]{cstype} } = [ $i ];
          }
      }

      for my $key (keys %types) {
         for my $i (sort {$a <=> $b} @{ $types{ $key } }) {
print $out qq|
                    case $i:|;
         }
print $out qq|
                    {
                        $key a = Convert.To$key( ((ListViewItem)x).SubItems[col].Text );
                        $key b = Convert.To$key( ((ListViewItem)y).SubItems[col].Text );
                        return sort == SortOrder.Ascending ? a.CompareTo(b) : b.CompareTo(a);
                    }|;
      }
print $out qq|
                }
                return 0;
            }
        }
    }
}

|;

      close $out;
   }
}


#################################################################################################
## Makes DB proxy objects
#################################################################################################
sub make_db_proxy_objects {
   my ($objects) = @_;

   foreach $object (@{$objects}){

      open my $out,'>gen/'.$object->{name}.'DB.cs' or die $!;

      print $out qq{/*
 * $object->{name}DB.cs
 *
 * THIS FILE IS AUTO-GENERATED FROM metadata.xml.
 * DO NOT UPDATE THIS FILE.
 * ANY CHANGES TO THIS FILE WILL BE LOST!!!
 * ALL CHANGES MUST BE MADE TO metadata.xml.
 */\n\n};

      print $out qq{
using System;
using System.Data;
using System.Data.SqlClient;
using System.Collections;

};

      print $out "namespace SAMPLEAPP {\n\n";

      print $out qq{\t/**
\t * $object->{name}DB
\t * A database proxy class for $object->{name} class.
\t *
\t * This is the stateless (static) service provider class
\t * to create, update, retrieve, and delete $object->{name}
\t * objects.
\t *
\t * This class uses SqlClient package to work with SQL Server.
\t *
\t * Application settings need to have "dbconnectstr" key defined!!!
\t *
\t * You never need to create an instance of this class.
\t */\n};

      print $out "\tpublic class $object->{name}DB\n\t{\n";

      #create; connection, transaction
      print $out "\t\t///<summary>This method creates a new $object->{name} object in database; \n";
      print $out "\t\t/// using a given database connection, and a started transaction.</summary>\n";
      print $out "\t\tpublic static Int32 Create".$object->{name}."(".$object->{name}." o, SqlConnection conn, SqlTransaction trans)\n\t\t{\n";
      print $out "\t\t\tSqlCommand cmd = new SqlCommand(\"Create".$object->{name}."\",conn);\n";
      print $out "\t\t\tcmd.CommandType = CommandType.StoredProcedure;\n";
      print $out "\t\t\tSqlParameter return_value = cmd.Parameters.Add(\"RETURN_VALUE\", SqlDbType.Int);\n";
      print $out "\t\t\treturn_value.Direction = ParameterDirection.ReturnValue;\n";
      foreach $attrib (@{$object->{attrib}}){
         if(!$attrib->{calculated} && !$attrib->{auto}){
            print $out "\t\t\tSqlParameter param_$attrib->{name} = cmd.Parameters.Add(\"\@".$attrib->{name}."\", $attrib->{csdbtype});\n";

            print $out "\t\t\tif(((object)o.$attrib->{cspname}) == null){\n";
            if($attrib->{required}){
               print $out "\t\t\t\tthrow new ApplicationException(\"$attrib->{fname} in record $object->{fname} is required.\");\n";
            }else{
               print $out "\t\t\t\tparam_$attrib->{name}.Value = DBNull.Value;\n";
            }
            print $out "\t\t\t}else{\n";
            print $out "\t\t\t\tparam_$attrib->{name}.Value = o.$attrib->{cspname};\n";
            print $out "\t\t\t}\n";

         }
      }
      print $out "\t\t\tif(trans != null)\n";
      print $out "\t\t\t\tcmd.Transaction = trans;\n";
      print $out "\t\t\tcmd.ExecuteNonQuery();\n";
      print $out "\t\t\treturn (Int32)cmd.Parameters[\"RETURN_VALUE\"].Value;\n";
      print $out "\t\t}\n\n";

      #create
      print $out "\t\t///<summary>This method creates a new $object->{name} object in database.</summary>\n";
      print $out "\t\tpublic static Int32 Create".$object->{name}."(".$object->{name}." o)\n\t\t{\n";
      print $out "\t\t\tInt32 rval = 0;\n";
      print $out "\t\t\tSqlConnection conn = DbConnectionHelper.GetConnection();\n";
      print $out "\t\t\tconn.Open();\n";
      print $out "\t\t\tSqlTransaction trans = conn.BeginTransaction();\n";
      print $out "\t\t\ttry{\n";
      print $out "\t\t\t\trval = Create".$object->{name}."(o,conn,trans);\n";
      print $out "\t\t\t\ttrans.Commit();\n";
      print $out "\t\t\t}catch(Exception e){\n";
      print $out "\t\t\t\ttrans.Rollback();\n";
      print $out "\t\t\t\tthrow new ApplicationException(\"Problem creating a $object->{name} record in the database:\"+e.Message);\n";
      print $out "\t\t\t}finally{\n";
      print $out "\t\t\t\tconn.Close();\n";
      print $out "\t\t\t}\n";
      print $out "\t\t\treturn rval;\n";
      print $out "\t\t}\n\n";

      # update
      print $out "\t\t///<summary>This method updates $object->{name} object in database.</summary>\n";
      print $out "\t\tpublic static void Update".$object->{name}."(".$object->{name}." o, SqlConnection conn, SqlTransaction trans)\n\t\t{\n";
      print $out "\t\t\tSqlCommand cmd = new SqlCommand(\"Update".$object->{name}."\",conn);\n";
      print $out "\t\t\tcmd.CommandType = CommandType.StoredProcedure;\n";
      foreach $attrib (@{$object->{attrib}}){
         if(!$attrib->{calculated}){
            print $out "\t\t\tSqlParameter param_$attrib->{name} = cmd.Parameters.Add(\"\@".$attrib->{name}."\", $attrib->{csdbtype});\n";
            print $out "\t\t\tif(((object)o.$attrib->{cspname}) == null){\n";

            if($attrib->{required}){
               print $out "\t\t\t\tthrow new ApplicationException(\"$attrib->{fname} in record $object->{fname} is required.\");\n";
            }else{
               print $out "\t\t\t\tparam_$attrib->{name}.Value = DBNull.Value;\n";
            }
            print $out "\t\t\t}else{\n";
            print $out "\t\t\t\tparam_$attrib->{name}.Value = o.$attrib->{cspname};\n";
            print $out "\t\t\t}\n";
         }
      }
      print $out "\t\t\tif(trans != null)\n";
      print $out "\t\t\t\tcmd.Transaction = trans;\n";
      print $out "\t\t\tcmd.ExecuteNonQuery();\n";
      print $out "\t\t}\n\n";

      # update
      print $out "\t\t///<summary>This method updates $object->{name} object in database.</summary>\n";
      print $out "\t\tpublic static void Update".$object->{name}."(".$object->{name}." o)\n\t\t{\n";
      print $out "\t\t\tSqlConnection conn = DbConnectionHelper.GetConnection();\n";
      print $out "\t\t\tconn.Open();\n";
      print $out "\t\t\tSqlTransaction trans = conn.BeginTransaction();\n";
      print $out "\t\t\ttry{\n";
      print $out "\t\t\t\tUpdate".$object->{name}."(o,conn,trans);\n";
      print $out "\t\t\t\ttrans.Commit();\n";
      print $out "\t\t\t}catch(Exception e){\n";
      print $out "\t\t\t\ttrans.Rollback();\n";
      print $out "\t\t\t\tthrow new ApplicationException(\"Problem updating $object->{name} record in the database:\"+e.Message);\n";
      print $out "\t\t\t}finally{\n";
      print $out "\t\t\t\tconn.Close();\n";
      print $out "\t\t\t}\n";
      print $out "\t\t}\n\n";

      #delete
      print $out "\t\t///<summary>This method deletes $object->{name} object from the database.</summary>\n";
      print $out "\t\tpublic static void Delete".$object->{name}."(";

      print $out ${$object->{attrib_names}}{$object->{firstkey}}->{cstype}.' pk';
      print $out ", SqlConnection conn, SqlTransaction trans)\n\t\t{\n";
      print $out "\t\t\tSqlCommand cmd = new SqlCommand(\"Delete".$object->{name}."\",conn);\n";
      print $out "\t\t\tcmd.CommandType = CommandType.StoredProcedure;\n";

      print $out "\t\t\tSqlParameter param_".
         ${$object->{attrib_names}}{$object->{firstkey}}->{name}.
         " = cmd.Parameters.Add(\"\@".
         ${$object->{attrib_names}}{$object->{firstkey}}->{name}.
         "\", ".${$object->{attrib_names}}{$object->{firstkey}}->{csdbtype}.");\n";
      print $out "\t\t\tparam_".
         ${$object->{attrib_names}}{$object->{firstkey}}->{name}.
         ".Value = pk;\n";
      print $out "\t\t\tif(trans != null)\n";
      print $out "\t\t\t\tcmd.Transaction = trans;\n";
      print $out "\t\t\tcmd.ExecuteNonQuery();\n";
      print $out "\t\t}\n\n";

      #delete
      print $out "\t\t///<summary>This method deletes $object->{name} object from the database.</summary>\n";
      print $out "\t\tpublic static void Delete".$object->{name}."(";
      print $out ${$object->{attrib_names}}{$object->{firstkey}}->{cstype}.' pk';
      print $out ")\n\t\t{\n";
      print $out "\t\t\tSqlConnection conn = DbConnectionHelper.GetConnection();\n";
      print $out "\t\t\tconn.Open();\n";
      print $out "\t\t\tSqlTransaction trans = conn.BeginTransaction();\n";
      print $out "\t\t\ttry{\n";
      print $out "\t\t\t\tDelete".$object->{name}."(pk,conn,trans);\n";
      print $out "\t\t\t\ttrans.Commit();\n";
      print $out "\t\t\t}catch(Exception e){\n";
      print $out "\t\t\t\ttrans.Rollback();\n";
      print $out "\t\t\t\tthrow new ApplicationException(\"Problem deleting $object->{name} record from the database:\"+e.Message);\n";
      print $out "\t\t\t}finally{\n";
      print $out "\t\t\t\tconn.Close();\n";
      print $out "\t\t\t}\n";
      print $out "\t\t}\n\n";


      # get
      print $out "\t\t///<summary>This method gets the $object->{name} object from the database by primary key.</summary>\n";
      print $out "\t\tpublic static $object->{name} Get".$object->{name}."(";

      print $out ${$object->{attrib_names}}{$object->{firstkey}}->{cstype}.' '.
         ${$object->{attrib_names}}{$object->{firstkey}}->{csname};
      print $out ")\n\t\t{\n";
      print $out "\t\t\tSqlConnection conn = DbConnectionHelper.GetConnection();\n";
      print $out "\t\t\tSqlCommand cmd = new SqlCommand(\"Get".$object->{name}."\",conn);\n";
      print $out "\t\t\tcmd.CommandType = CommandType.StoredProcedure;\n";

      print $out "\t\t\tSqlParameter param_".
         ${$object->{attrib_names}}{$object->{firstkey}}->{name}.
         " = cmd.Parameters.Add(\"\@".
         ${$object->{attrib_names}}{$object->{firstkey}}->{name}.
         "\", ".${$object->{attrib_names}}{$object->{firstkey}}->{csdbtype}.");\n";
      print $out "\t\t\tparam_".
         ${$object->{attrib_names}}{$object->{firstkey}}->{name}.
         ".Value = ".${$object->{attrib_names}}{$object->{firstkey}}->{csname}.";\n";

      print $out "\t\t\tconn.Open();\n";
      print $out "\t\t\tSqlDataReader dbreader = cmd.ExecuteReader();\n";
      print $out "\t\t\t".$object->{name}." o = new ".$object->{name}."();\n";
      print $out "\t\t\tif (dbreader.Read()){\n";
      my $i = 0;
      foreach $attrib (@{$object->{attrib}}){
          if(!$attrib->{calculated}){
             print $out "\t\t\t\tif(!dbreader.IsDBNull($i)){\n";
             print $out "\t\t\t\t\to.".$attrib->{cspname}." = dbreader.Get".$attrib->{cstype}."($i);\n";
             print $out "\t\t\t\t}\n";
          }
          $i++;
      }
      print $out "\t\t\t}else{\n";
      print $out "\t\t\t\to = null;\n";
      print $out "\t\t\t}\n";
      print $out "\t\t\tdbreader.Close();\n";
      print $out "\t\t\tconn.Close();\n";
      print $out "\t\t\treturn o;\n";
      print $out "\t\t}\n\n";



      # get all
      print $out "\t\t///<summary>This method gets ALL the $object->{name} objects from the database.</summary>\n";
      print $out "\t\tpublic static ArrayList GetAll".$object->{name}."s()\n\t\t{\n";
      print $out "\t\t\tSqlConnection conn = DbConnectionHelper.GetConnection();\n";
      print $out "\t\t\tSqlCommand cmd = new SqlCommand(\"GetAll".$object->{name}."s\",conn);\n";
      print $out "\t\t\tcmd.CommandType = CommandType.StoredProcedure;\n";
      print $out "\t\t\tconn.Open();\n";
      print $out "\t\t\tSqlDataReader dbreader = cmd.ExecuteReader();\n";
      print $out "\t\t\tArrayList arr = new ArrayList();\n";
      print $out "\t\t\twhile (dbreader.Read()){\n";
      print $out "\t\t\t\t".$object->{name}." o = new ".$object->{name}."();\n";
      my $i = 0;
      foreach $attrib (@{$object->{attrib}}){
          if(!$attrib->{calculated}){
             print $out "\t\t\t\tif(!dbreader.IsDBNull($i)){\n";
             print $out "\t\t\t\t\to.".$attrib->{cspname}." = dbreader.Get".$attrib->{cstype}."($i);\n";
             if($attrib->{password}){
                print $out "\t\t\t\t\to.".$attrib->{cspname}." = \"*****\";\n";
             }
             print $out "\t\t\t\t}\n";
          }
          $i++;
      }
      print $out "\t\t\t\tarr.Add(o);\n";
      print $out "\t\t\t}\n";
      print $out "\t\t\tdbreader.Close();\n";
      print $out "\t\t\tconn.Close();\n";
      print $out "\t\t\treturn arr;\n";
      print $out "\t\t}\n\n";

      # work on the queries
      foreach my $query (@{$object->{query}}){
          print $out "\t\t///<summary>$query->{description}</summary>\n";
          if($query->{single}){     # returning only 1 record?


               # TODO :-)

          }else{        # returning many records

             # get many via query
             print $out "\t\tpublic static ArrayList Get".$query->{name}."(";

             splice @str;
             foreach $param (@{$query->{param}}){
                push @str,$param->{cstype}.' '.$param->{name};
             }
             print $out join ",",@str;

             print $out ")\n\t\t{\n";
             print $out "\t\t\tSqlConnection conn = DbConnectionHelper.GetConnection();\n";
             print $out "\t\t\tSqlCommand cmd = new SqlCommand(\"Get".$query->{name}."\",conn);\n";
             print $out "\t\t\tcmd.CommandType = CommandType.StoredProcedure;\n";

             foreach $param (@{$query->{param}}){
                print $out "\t\t\tSqlParameter param_$param->{name} = cmd.Parameters.Add(\"\@".$param->{name}."\", $param->{csdbtype});\n";
                print $out "\t\t\tparam_$param->{name}.Value = $param->{name};\n";
             }
             print $out "\t\t\tconn.Open();\n";
             print $out "\t\t\tSqlDataReader dbreader = cmd.ExecuteReader();\n";
             print $out "\t\t\tArrayList arr = new ArrayList();\n";
             print $out "\t\t\twhile (dbreader.Read()){\n";
             print $out "\t\t\t\t".$object->{name}." o = new ".$object->{name}."();\n";
             my $i = 0;
            foreach $attrib (@{$object->{attrib}}){
               if(!$attrib->{calculated}){
                  print $out "\t\t\t\tif(!dbreader.IsDBNull($i)){\n";
                  print $out "\t\t\t\t\to.".$attrib->{cspname}." = dbreader.Get".$attrib->{cstype}."($i);\n";
                  print $out "\t\t\t\t}\n";
               }
               $i++;
            }
             print $out "\t\t\t\tarr.Add(o);\n";
             print $out "\t\t\t}\n";
             print $out "\t\t\tdbreader.Close();\n";
             print $out "\t\t\tconn.Close();\n";
             print $out "\t\t\treturn arr;\n";
             print $out "\t\t}\n\n";
          }
      }

      # work on the actions
      foreach my $query (@{$object->{action}}){
          print $out "\t\t///<summary>$query->{description}</summary>\n";

          # get many via query
          print $out "\t\tpublic static void Do".$query->{name}."(";

          splice @str;
          foreach $param (@{$query->{param}}){
             push @str,$param->{cstype}.' '.$param->{name};
          }
          print $out join ",",@str;

          print $out ", SqlConnection conn, SqlTransaction trans)\n\t\t{\n";
          print $out "\t\t\tSqlCommand cmd = new SqlCommand(\"Do".$query->{name}."\",conn);\n";
          print $out "\t\t\tcmd.CommandType = CommandType.StoredProcedure;\n";
          foreach $param (@{$query->{param}}){
             print $out "\t\t\tSqlParameter param_$param->{name} = cmd.Parameters.Add(\"\@".$param->{name}."\", $param->{csdbtype});\n";
             print $out "\t\t\tparam_$param->{name}.Value = $param->{name};\n";
          }
          print $out "\t\t\tif(trans != null)\n";
          print $out "\t\t\t\tcmd.Transaction = trans;\n";
          print $out "\t\t\tcmd.ExecuteNonQuery();\n";
          print $out "\t\t}\n\n";
      }


      # work on the actions
      foreach my $query (@{$object->{action}}){
          print $out "\t\t///<summary>$query->{description}</summary>\n";

          # get many via query
          print $out "\t\tpublic static void Do".$query->{name}."(";

          splice @str;
          foreach $param (@{$query->{param}}){
             push @str,$param->{cstype}.' '.$param->{name};
          }
          print $out join ",",@str;

          print $out ")\n\t\t{\n";
          print $out "\t\t\tSqlConnection conn = DbConnectionHelper.GetConnection();\n";
          print $out "\t\t\tconn.Open();\n";
          print $out "\t\t\tSqlTransaction trans = conn.BeginTransaction();\n";
          print $out "\t\t\ttry{\n";
          print $out "\t\t\t\tDo".$query->{name}."(";
          splice @str;
          foreach $param (@{$query->{param}}){
             push @str,$param->{name};
          }
          print $out join ",",@str;
          print $out ",conn,trans);\n";
          print $out "\t\t\t\ttrans.Commit();\n";
          print $out "\t\t\t}catch(Exception e){\n";
          print $out "\t\t\t\ttrans.Rollback();\n";
          print $out "\t\t\t\tthrow new ApplicationException(\"Problem Performing Action $query->{name}: \"+e.Message);\n";
          print $out "\t\t\t}finally{\n";
          print $out "\t\t\t\tconn.Close();\n";
          print $out "\t\t\t}\n";
          print $out "\t\t}\n\n";
      }

      # work on the multi-queries
      foreach my $query (@{$object->{multiquery}}){
          print $out "\t\t///<summary>$query->{description}</summary>\n";


          # get many via query
          print $out "\t\tpublic static ArrayList Get".$query->{name}."(";

          splice @str;
          foreach $param (@{$query->{param}}){
             push @str,$param->{cstype}.' '.$param->{name};
          }
          print $out join ",",@str;

          print $out ")\n\t\t{\n";
          print $out "\t\t\tSqlConnection conn = DbConnectionHelper.GetConnection();\n";
          print $out "\t\t\tSqlCommand cmd = new SqlCommand(\"Get".$query->{name}."\",conn);\n";
          print $out "\t\t\tcmd.CommandType = CommandType.StoredProcedure;\n";
          foreach $param (@{$query->{param}}){
             print $out "\t\t\tSqlParameter param_$param->{name} = cmd.Parameters.Add(\"\@".$param->{name}."\", $param->{csdbtype});\n";
             print $out "\t\t\tparam_$param->{name}.Value = $param->{name};\n";
          }
          print $out "\t\t\tconn.Open();\n";
          print $out "\t\t\tSqlDataReader dbreader = cmd.ExecuteReader();\n";
          print $out "\t\t\tArrayList arr = new ArrayList();\n";
          print $out "\t\t\twhile (dbreader.Read()){\n";
          my $i = 0;
          my $j = 1;

          my @constructor_params = ();
          for my $queries (@{$query->{queries}}) {
             # get the object reference from the global object hash.
             my $object = $objects{$queries};
             print $out "\t\t\t\t".$object->{name}." o$j = new ".$object->{name}."();\n";
             push @constructor_params,"o$j";
             foreach $attrib (@{$object->{attrib}}){
                 if(!$attrib->{calculated}){
                    print $out "\t\t\t\tif(!dbreader.IsDBNull($i)){\n";
                    print $out "\t\t\t\t\to$j.".$attrib->{cspname}." = dbreader.Get".$attrib->{cstype}."($i);\n";
                    print $out "\t\t\t\t}\n";
                 }
                 $i++;
             }
             $j++;
             print $out "\n";
          }
          print $out "\t\t\t\t".$query->{returns}." o = new ".$query->{returns}."(";
          print $out join ',',@constructor_params;
          print $out ");\n";
          print $out "\t\t\t\tarr.Add(o);\n";
          print $out "\t\t\t}\n";
          print $out "\t\t\tdbreader.Close();\n";
          print $out "\t\t\tconn.Close();\n";
          print $out "\t\t\treturn arr;\n";
          print $out "\t\t}\n\n";
      }

      # work on the queries
      foreach my $query (@{$object->{queryds}}){
          print $out "\t\t///<summary>$query->{description}</summary>\n";

          # get many via query
          print $out "\t\tpublic static DataSet Get".$query->{name}."(";

          splice @str;
          foreach $param (@{$query->{param}}){
             push @str,$param->{cstype}.' '.$param->{name};
          }
          print $out join ",",@str;

          print $out ")\n\t\t{\n";
          print $out "\t\t\tDataSet dataSet = new DataSet();\n";
          print $out "\t\t\tSqlConnection conn = DbConnectionHelper.GetConnection();\n";
          print $out "\t\t\tSqlCommand cmd = new SqlCommand(\"Get".$query->{name}."\",conn);\n";
          print $out "\t\t\tcmd.CommandType = CommandType.StoredProcedure;\n";
          foreach $param (@{$query->{param}}){
             print $out "\t\t\tSqlParameter param_$param->{name} = cmd.Parameters.Add(\"\@".$param->{name}."\", $param->{csdbtype});\n";
             print $out "\t\t\tparam_$param->{name}.Value = $param->{name};\n";
          }
          print $out "\t\t\tSqlDataAdapter dataAdapter = new SqlDataAdapter();\n";
          print $out "\t\t\tdataAdapter.SelectCommand = cmd;\n";
          print $out "\t\t\tconn.Open();\n";
          print $out "\t\t\tdataAdapter.Fill(dataSet);\n";
          print $out "\t\t\tconn.Close();\n";
          print $out "\t\t\treturn dataSet;\n";
          print $out "\t\t}\n\n";
      }



      # work on the squerie
      foreach my $query (@{$object->{squery}}){
          print $out "\t\t///<summary>$query->{description}</summary>\n";
          # get many via query
          print $out "\t\tpublic static $query->{returns}{cstype} Get".$query->{name}."(";
          splice @str;
          foreach $param (@{$query->{param}}){
             push @str,$param->{cstype}.' '.$param->{name};
          }
          print $out join ",",@str;
          print $out ")\n\t\t{\n";
          print $out "\t\t\tSqlConnection conn = DbConnectionHelper.GetConnection();\n";
          print $out "\t\t\tSqlCommand cmd = new SqlCommand(\"Get".$query->{name}."\",conn);\n";
          print $out "\t\t\tcmd.CommandType = CommandType.StoredProcedure;\n";
          foreach $param (@{$query->{param}}){
             print $out "\t\t\tSqlParameter param_$param->{name} = cmd.Parameters.Add(\"\@".$param->{name}."\", $param->{csdbtype});\n";
             print $out "\t\t\tparam_$param->{name}.Value = $param->{name};\n";
          }
          print $out "\t\t\tconn.Open();\n";
          print $out "\t\t\t$query->{returns}{cstype} n = ($query->{returns}{cstype})cmd.ExecuteScalar();\n";
          print $out "\t\t\tconn.Close();\n";
          print $out "\t\t\treturn n;\n";
          print $out "\t\t}\n\n";
      }

      # get count
      print $out "\t\t///<summary>This method gets the count of $object->{name} objects in the database.</summary>\n";
      print $out "\t\tpublic static Int32 Get".$object->{name}."Count()\n\t\t{\n";
      print $out "\t\t\tSqlConnection conn = DbConnectionHelper.GetConnection();\n";
      print $out "\t\t\tSqlCommand cmd = new SqlCommand(\"Get".$object->{name}."Count\",conn);\n";
      print $out "\t\t\tcmd.CommandType = CommandType.StoredProcedure;\n";
      print $out "\t\t\tconn.Open();\n";
      print $out "\t\t\tInt32 n = (Int32)cmd.ExecuteScalar();\n";
      print $out "\t\t\tconn.Close();\n";
      print $out "\t\t\treturn n;\n";
      print $out "\t\t}\n\n";

      print $out "\t}\n";
      print $out "\n}\n";

      close $out;
   }
}












#################################################################################################
## make value objects
#################################################################################################
sub make_value_objects {
   my ($objects) = @_;

   foreach $object (@{$objects}){

      open my $out,'>gen/'.$object->{name}.'.cs' or die $!;

      print $out qq{/*
 * $object->{name}.cs
 *
 * THIS FILE IS AUTO-GENERATED FROM metadata.xml.
 * DO NOT UPDATE THIS FILE.
 * ANY CHANGES TO THIS FILE WILL BE LOST!!!
 * ALL CHANGES MUST BE MADE TO metadata.xml.
 */\n\n};

      print $out qq{
using System;
using System.Runtime.Serialization;

};

      print $out "namespace SAMPLEAPP {\n\n";

      print $out qq{\t/**
\t * $object->{name}
\t * $object->{description}
\t *
\t * This is a Serializable VALUE object. A structure. It has no
\t * functionality on its own. It is a data carrier class.
\t */\n};

      print $out qq{\t[Serializable]\n};
      print $out "\tpublic class $object->{name}\n\t{\n";

      foreach $attrib (@{$object->{attrib}}){
         print $out "\t\tprivate ".$attrib->{cstype}." ".$attrib->{csname}.";\n";
      }

      print $out "\n";
      foreach $attrib (@{$object->{attrib}}){
         print $out "\t\tpublic ".$attrib->{cstype}." ".$attrib->{cspname}."\n";
         print $out "\t\t{\n";
         print $out "\t\t\tget { return this.$attrib->{csname}; }\n";
         print $out "\t\t\tset { this.$attrib->{csname} = value; }\n";
         print $out "\t\t}\n\n";
      }

      print $out "\t}\n";
      print $out "\n}\n";

      close $out;
   }
}















#################################################################################################
## make SQL
#################################################################################################
sub make_sql {
    my ($objects) = @_;

    open my $out,">dbgen.sql" or die $!;

    print $out "-- This file is auto GENERATED from metadata.xml file.\n";
    print $out "-- DO NOT MODIFY THIS FILE!!!\n";
    print $out "-- All changes should be made to metadata.xml instead.\n\n\n";

    foreach $object (@{$objects}){
        print $out "-- ".$object->{description}."\n";

        print $out "IF EXISTS(SELECT * FROM sysobjects WHERE (name='".$object->{name}."') AND (xtype='U'))\n\tDROP TABLE ".$object->{sqlname}."\nGO\n\n";

        print $out "CREATE TABLE ".$object->{sqlname}." (\n";

        foreach $attrib (@{$object->{attrib}}){
            if(!$attrib->{calculated}){
                print $out "\t".$attrib->{sqlname}." ".$attrib->{sqltype};
                print $out " DEFAULT($attr->{sqldefault})" if $attr->{sqldefault};
                print $out " NOT NULL" if $attrib->{required};
                print $out " IDENTITY(1,1)" if $attrib->{auto};
                print $out ",\n";
            }
        }
        print $out "\tPRIMARY KEY(".$object->{sqlpkey}.")";
        foreach $key (@{$object->{key}}){
            my $attrib = $object->{attrib_names}{$key};
            print $out ",\n\tKEY (".$attrib->{sqlname}.")";
        }
        foreach $unique (@{$object->{unique}}){
            my $attrib = $object->{attrib_names}{$unique};
            print $out ",\n\tUNIQUE (".$attrib->{sqlname}.")";
        }

        print $out "\n);\n\nGO\n\n";


# Alternate way of setting a primary key
#        print $out "CREATE UNIQUE CLUSTERED INDEX PK_".$object->{name}." ON ".$object->{sqlname}." (".$object->{sqlpkey}.")";
#        print $out "\n\nGO\n\n";

        print $out "-- Stored Procedures to work with the table.\n";

        print $out "\n-- Creates a new $object->{name} record entry.\n";

        print $out "IF EXISTS(SELECT * FROM sysobjects WHERE (name='Create".$object->{name}."') AND (xtype='P'))\n\tDROP PROCEDURE Create".$object->{name}."\nGO\n\n";

        print $out "CREATE PROCEDURE Create".$object->{name}."\n(\n";
        splice @str;
        foreach $attrib (@{$object->{attrib}}){
            if(!$attrib->{calculated}){
                push (@str,"\t\@".$attrib->{name}." ".$attrib->{sqltype}) if !$attrib->{auto};
            }
        }
        print $out join ",\n",@str;
        print $out "\n) \nAS \n";
        print $out "INSERT INTO ".$object->{sqlname}." (\n";
        splice @str;
        foreach $attrib (@{$object->{attrib}}){
            if(!$attrib->{calculated}){
                push (@str,"\t".$attrib->{sqlname}) if !$attrib->{auto};
            }
        }
        print $out join ",\n",@str;
        print $out "\n) \nVALUES\n(\n";
        splice @str;
        foreach $attrib (@{$object->{attrib}}){
            if(!$attrib->{calculated}){
                push (@str,"\t\@".$attrib->{name}) if !$attrib->{auto};
            }
        }
        print $out join ",\n",@str;
        print $out "\n)\n";

        foreach $attrib (@{$object->{attrib}}){
            if(!$attrib->{calculated}){
                print $out 'RETURN @@IDENTITY '."\n" if $attrib->{auto};
            }
        }
        print $out "\nGO\n\n";




        print $out "\n-- Update a $object->{name} record in the database.\n";

        print $out "IF EXISTS(SELECT * FROM sysobjects WHERE (name='Update".$object->{name}."') AND (xtype='P'))\n\tDROP PROCEDURE Update".$object->{name}."\nGO\n\n";

        print $out "CREATE PROCEDURE Update".$object->{name}."\n(\n";
        splice @str;
        foreach $attrib (@{$object->{attrib}}){
            if(!$attrib->{calculated}){
                push (@str,"\t\@".$attrib->{name}." ".$attrib->{sqltype});
            }
        }
        print $out join ",\n",@str;
        print $out "\n) \nAS \n";

        print $out "\tUPDATE ".$object->{sqlname}."\n\tSET\n";
        splice @str;
        foreach $attrib (@{$object->{attrib}}){
            if(!$attrib->{calculated}){
                push (@str,"\t".$attrib->{sqlname}.'=@'.$attrib->{name}) if !$attrib->{pkey};
            }
        }
        print $out join ",\n",@str;
        print $out "\n\tWHERE ";
        print $out ${$object->{attrib_names}}{$object->{firstkey}}->{sqlname}.'='.
            '@'.${$object->{attrib_names}}{$object->{firstkey}}->{name};

        print $out "\nGO\n\n";


        print $out "\n-- Delete a $object->{name} record from the database.\n";

        print $out "IF EXISTS(SELECT * FROM sysobjects WHERE (name='Delete".$object->{name}."') AND (xtype='P'))\n\tDROP PROCEDURE Delete".$object->{name}."\nGO\n\n";

        print $out "CREATE PROCEDURE Delete".$object->{name}."\n(\n";

        print $out "\t@".${$object->{attrib_names}}{$object->{firstkey}}->{name}.' '.
            ${$object->{attrib_names}}{$object->{firstkey}}->{sqltype};
        print $out "\n) \nAS \n";

        print $out "\tDELETE FROM ".$object->{sqlname}."\n";
        print $out "\tWHERE ";
        print $out ${$object->{attrib_names}}{$object->{firstkey}}->{sqlname}.'='.
            '@'.${$object->{attrib_names}}{$object->{firstkey}}->{name};

        print $out "\nGO\n\n";

        # get a single record.
        print $out "\n-- Get a $object->{name} record by primary key.\n";

        print $out "IF EXISTS(SELECT * FROM sysobjects WHERE (name='Get".$object->{name}."') AND (xtype='P'))\n\tDROP PROCEDURE Get".$object->{name}."\nGO\n\n";

        print $out "CREATE PROCEDURE Get".$object->{name}."\n(\n";

        print $out "\t@".${$object->{attrib_names}}{$object->{firstkey}}->{name}.' '.
            ${$object->{attrib_names}}{$object->{firstkey}}->{sqltype};
        print $out "\n) \nAS \n";

        print $out "\tSELECT * \n\tFROM ".$object->{sqlname}."\n";
        print $out "\tWHERE ";
        print $out ${$object->{attrib_names}}{$object->{firstkey}}->{sqlname}.'='.
            '@'.${$object->{attrib_names}}{$object->{firstkey}}->{name};

        print $out "\nGO\n\n";

        # get all records (top 1024).
        print $out "\n-- Get 'all' $object->{name} records; it actually gets TOP 256 records.\n";
        print $out "IF EXISTS(SELECT * FROM sysobjects WHERE (name='GetAll".$object->{name}."s') AND (xtype='P'))\n\tDROP PROCEDURE GetAll".$object->{name}."s\nGO\n\n";
        print $out "CREATE PROCEDURE GetAll".$object->{name}."s\n";
        print $out "AS\n";
        print $out "\tSELECT TOP 1024 * \n\tFROM ".$object->{sqlname}."\n";
        print $out "\nGO\n\n";

        # get count of all records.
        print $out "\n-- Get Count of all $object->{name} records.\n";
        print $out "IF EXISTS(SELECT * FROM sysobjects WHERE (name='Get".$object->{name}."Count') AND (xtype='P'))\n\tDROP PROCEDURE Get".$object->{name}."Count\nGO\n\n";
        print $out "CREATE PROCEDURE Get".$object->{name}."Count\n";
        print $out "AS\n";
        print $out "\tSELECT COUNT(*) \n\tFROM ".$object->{sqlname}."\n";
        print $out "\nGO\n\n";

        foreach my $query (@{$object->{query}},@{$object->{multiquery}},@{$object->{queryds}},@{$object->{squery}}){
            print $out "-- ".$query->{description}."\n";

            print $out "IF EXISTS(SELECT * FROM sysobjects WHERE (name='Get".$query->{name}."') AND (xtype='P'))\n\tDROP PROCEDURE Get".$query->{name}."\nGO\n\n";

            print $out "CREATE PROCEDURE Get".$query->{name}."\n(\n";

            splice @str;
            foreach $param (@{$query->{param}}){
               push (@str,"\t\@".$param->{name}.' '.$param->{sqltype});
            }
            print $out join ",\n",@str;
            print $out "\n) \nAS \n";

            my $sql = $query->{sql};
            foreach $param (@{$query->{param}}){
               $sql =~ s|\$$param->{name}\$|\@$param->{name}|gs;
            }
            print $out "\t$sql\n\nGO\n\n";
        }

        foreach my $query (@{$object->{action}}){
            print $out "-- ".$query->{description}."\n";

            print $out "IF EXISTS(SELECT * FROM sysobjects WHERE (name='Do".$query->{name}."') AND (xtype='P'))\n\tDROP PROCEDURE Do".$query->{name}."\nGO\n\n";

            print $out "CREATE PROCEDURE Do".$query->{name}."\n(\n";

            splice @str;
            foreach $param (@{$query->{param}}){
               push (@str,"\t\@".$param->{name}.' '.$param->{sqltype});
            }
            print $out join ",\n",@str;
            print $out "\n) \nAS \n";

            my $sql = $query->{sql};
            foreach $param (@{$query->{param}}){
               $sql =~ s|\$$param->{name}\$|\@$param->{name}|gs;
            }
            print $out "\t$sql\n\nGO\n\n";
        }


    }

   close $out;
}



















#################################################################################################
## turns a single object from xml parse tree into a metadata object
#################################################################################################
sub make_object {

    my ($object) = @_;

    my $obj = {};

    $obj->{name} = queryXML($object,'object/%name');
    $obj->{description} = queryXML($object,'object/description/$');

    # field name
    $obj->{fname} = queryXML($object,'object/fname/$');
    $obj->{fname} = $obj->{name} if !$obj->{fname};

    # plural name
    $obj->{pname} = queryXML($object,'object/pname/$');
    $obj->{pname} = $obj->{fname}.'s' if !$obj->{pname};

    # Global.cs array reference (if there is any?)
    $obj->{globalref} = queryXML($object,'object/globalref/$');
    $obj->{globalref} = $obj->{name}.'s' if !$obj->{globalref};

    $obj->{sqlname} = '['.$obj->{name}.']';

    $obj->{pkey} = [];
    $obj->{key} = [];
    $obj->{unique} = [];

    $obj->{attrib} = [];
    $obj->{attrib_names} = {};

    my @sqlpkey = ();

    for $attrib (queryXML($object,'object/@attribute')){

        my $attr = {};

        $attr->{name} = queryXML($attrib,'attribute/%name');
        $attr->{type} = queryXML($attrib,'attribute/type/$');
        $attr->{size} = queryXML($attrib,'attribute/size/$');
        $attr->{oneof} = queryXML($attrib,'attribute/oneof/$');
        $attr->{str} = $gen_types->{ $attr->{type} }{str};
        $attr->{sqldefault} = queryXML($attrib,'attribute/sqldefault/$');
        $attr->{sqlname} = '['.$attr->{name}.']';
        $attr->{sqltype} = to_SQL_type($attr->{type},$attr->{size});
        $attr->{required} = 1 if queryXML($attrib,'attribute/@required');
        $attr->{auto} = 1 if queryXML($attrib,'attribute/@auto');
        $attr->{calculated} = 1 if queryXML($attrib,'attribute/@calculated');
        $attr->{pkey} = 1 if queryXML($attrib,'attribute/@pkey');
        $attr->{password} = 1 if queryXML($attrib,'attribute/@password');

        $attr->{csname} = $attr->{name};
        $attr->{cstype} = $gen_types->{$attr->{type}}{cs};
        $attr->{csdbtype} = $gen_types->{$attr->{type}}{csdb};
        $attr->{cspname} = ucfirst($attr->{name});

        $attr->{dispsize} = $gen_types->{$attr->{type}}{dispsize};
        $attr->{dispsize} *= $attr->{size} if $attr->{size};
        $attr->{dispsize} = 32 if $attr->{dispsize} < 32;

        # field name
        $attr->{fname} = queryXML($attrib,'attribute/fname/$');
        $attr->{fname} = $attr->{cspname} if !$attr->{fname};

        $attr->{description} = queryXML($attrib,'attribute/description/$');
        $attr->{description} = $attr->{fname} if !$attr->{description};


        push @{$obj->{pkey}},$attr->{name} if $attr->{pkey};
        push @{$obj->{auto}},$attr->{name} if $attr->{auto};
        push @{$obj->{key}},$attr->{name} if queryXML($attrib,'attribute/@key');
        push @{$obj->{unique}},$attr->{name} if queryXML($attrib,'attribute/@unique');
        push @sqlpkey,$attr->{sqlname} if queryXML($attrib,'attribute/@pkey');
        push @{$obj->{attrib}},$attr;
        ${$obj->{attrib_names}}{$attr->{name}} = $attr;
    }
    $obj->{sqlpkey} = join ',',@sqlpkey;
    $obj->{firstkey} = ${$obj->{pkey}}[0];
    $obj->{firstauto} = ${$obj->{auto}}[0];


#    <query name="PersonByName">
#        <description>Given a first name and last name, finds people that match'em.</description>
#        <parameter name="lname">
#            <type>string</type>
#            <size>32</size>
#        </parameter>
#        <parameter name="fname">
#            <type>string</type>
#            <size>32</size>
#        </parameter>
#        <sql>SELECT p.* FROM [Person] p WHERE p.[lname] = $lname$ AND p.[fname] = $fname$</sql>
#    </query>


    $obj->{query} = [];
    for $query (queryXML($object,'object/@query')){

        my $attr = {};

        $attr->{name} = queryXML($query,'query/%name');
        $attr->{description} = queryXML($query,'query/description/$');
        $attr->{single} = 1 if queryXML($query,'query/@single');

        $attr->{sql} = queryXML($query,'query/sql/$');

        $attr->{param} = [];
        for $param (queryXML($query,'query/@parameter')){
            my $parm = {};
            $parm->{name} = queryXML($param,'parameter/%name');
            $parm->{type} = queryXML($param,'parameter/type/$');
            $parm->{size} = queryXML($param,'parameter/size/$');
            $parm->{str} = $gen_types{ $parm->{type} }{str};
            $parm->{sqltype} = to_SQL_type($parm->{type},$parm->{size});
            $parm->{cstype} = $gen_types->{$parm->{type}}{cs};
            $parm->{csdbtype} = $gen_types->{$parm->{type}}{csdb};
            push @{$attr->{param}},$parm;
        }
        push @{$obj->{query}},$attr;
    }

#    <!-- preliminary support for actions -->
#    <action name="RemoveRolesForUser">
#        <description>Given a userid, removes all of UserRole objects that match the userid.</description>
#        <parameter name="userid">
#            <type>int</type>
#        </parameter>
#        <sql>DELETE FROM [UserRole] WHERE [userid] = $userid$</sql>
#    </action>

    $obj->{action} = [];
    for $query (queryXML($object,'object/@action')){

        my $attr = {};

        $attr->{name} = queryXML($query,'action/%name');
        $attr->{description} = queryXML($query,'action/description/$');
        $attr->{single} = 1 if queryXML($query,'action/@single');

        $attr->{sql} = queryXML($query,'action/sql/$');

        $attr->{param} = [];
        for $param (queryXML($query,'action/@parameter')){
            my $parm = {};
            $parm->{name} = queryXML($param,'parameter/%name');
            $parm->{type} = queryXML($param,'parameter/type/$');
            $parm->{size} = queryXML($param,'parameter/size/$');
            $parm->{str} = $gen_types{ $parm->{type} }{str};
            $parm->{sqltype} = to_SQL_type($parm->{type},$parm->{size});
            $parm->{cstype} = $gen_types->{$parm->{type}}{cs};
            $parm->{csdbtype} = $gen_types->{$parm->{type}}{csdb};
            push @{$attr->{param}},$parm;
        }
        push @{$obj->{action}},$attr;
    }

#    <!-- preliminary support for multi-queries; those that return
#     more than one record. These require the presense of a 'value' class there
#     the constructor (or one of the constructors) matches the order of the `queries'
#     variables. -->
#    <multiquery name="LicenseDetailsByPerson">
#        <description>Finds a license and the two people connected to it. Given one of the people.</description>
#        <parameter name="personid">
#            <type>int</type>
#        </parameter>
#        <queries>License</queries>
#        <queries>Person</queries>
#        <queries>Person</queries>
#        <returns>LicenseInfo</returns>
#        <sql>SELECT [l].*, [p1].*, [p2].* FROM [Person] [p1], [PersonLicense] [pl1], [License] [l], [PersonLicense] [pl2], [Person] [p2] WHERE [p1].[id] = $personid$ AND [p1].[id] = [pl1].[personid] AND [pl1].[licenseid] = [l].[id] AND [pl2].[licenseid] = [l].[id] AND [p2].[id] = [pl2].[personid] AND [p2].[id] != [p1].[id]</sql>
#    </multiquery>

    $obj->{multiquery} = [];
    for $query (queryXML($object,'object/@multiquery')){

        my $attr = {};

        $attr->{name} = queryXML($query,'multiquery/%name');
        $attr->{description} = queryXML($query,'multiquery/description/$');

        $attr->{sql} = queryXML($query,'multiquery/sql/$');
        $attr->{returns} = queryXML($query,'multiquery/returns/$');

        $attr->{param} = [];
        for $param (queryXML($query,'multiquery/@parameter')){
            my $parm = {};
            $parm->{name} = queryXML($param,'parameter/%name');
            $parm->{type} = queryXML($param,'parameter/type/$');
            $parm->{size} = queryXML($param,'parameter/size/$');
            $parm->{str} = $gen_types{ $parm->{type} }{str};
            $parm->{sqltype} = to_SQL_type($parm->{type},$parm->{size});
            $parm->{cstype} = $gen_types->{$parm->{type}}{cs};
            $parm->{csdbtype} = $gen_types->{$parm->{type}}{csdb};
            push @{$attr->{param}},$parm;
        }
        $attr->{queries} = [];
        for $param (queryXML($query,'multiquery/@queries')){
            my $blah;
            $blah = queryXML($param,'queries/$');
            push @{$attr->{queries}},$blah;
        }
        push @{$obj->{multiquery}},$attr;
    }

#
#    <queryds name="NewlyCreatedLicenses">
#       <description>Finds new Licenses for such and such</description>
#       <parameter name="groom">
#           <type>string</type>
#       </parameter>
#       <sql>SELECT * FROM [License] WHERE [bridefatherfname] = $groom$</sql>
#    </queryds>
    $obj->{queryds} = [];
    for $query (queryXML($object,'object/@queryds')){

        my $attr = {};

        $attr->{name} = queryXML($query,'queryds/%name');
        $attr->{description} = queryXML($query,'queryds/description/$');
        $attr->{sql} = queryXML($query,'queryds/sql/$');

        $attr->{param} = [];
        for $param (queryXML($query,'queryds/@parameter')){
            my $parm = {};
            $parm->{name} = queryXML($param,'parameter/%name');
            $parm->{type} = queryXML($param,'parameter/type/$');
            $parm->{size} = queryXML($param,'parameter/size/$');
            $parm->{str} = $gen_types{ $parm->{type} }{str};
            $parm->{sqltype} = to_SQL_type($parm->{type},$parm->{size});
            $parm->{cstype} = $gen_types->{$parm->{type}}{cs};
            $parm->{csdbtype} = $gen_types->{$parm->{type}}{csdb};
            push @{$attr->{param}},$parm;
        }
        push @{$obj->{queryds}},$attr;
    }





#    <squery name="GetSequence">
#        <description>Get a single value.</description>
#        <parameter name="name">
#            <type>string</type>
#            <size>32</size>
#        </parameter>
#        <returns>int</returns>
#        <sql>SELECT [value] FROM [Sequence] WHERE [name]=$name$</sql>
#    </squery>

    $obj->{squery} = [];
    for $query (queryXML($object,'object/@squery')){

        my $attr = {};

        $attr->{name} = queryXML($query,'squery/%name');
        $attr->{description} = queryXML($query,'squery/description/$');

        $attr->{sql} = queryXML($query,'squery/sql/$');
        my $returns = queryXML($query,'squery/returns/$');
        $attr->{returns} = {
            str => $gen_types{$returns}{str},
            sqltype => $gen_types{$returns}{sql},
            cstype => $gen_types->{$returns}{cs},
            csdbtype => $gen_types->{$returns}{csdb},
        };

        $attr->{param} = [];
        for $param (queryXML($query,'squery/@parameter')){
            my $parm = {};
            $parm->{name} = queryXML($param,'parameter/%name');
            $parm->{type} = queryXML($param,'parameter/type/$');
            $parm->{size} = queryXML($param,'parameter/size/$');
            $parm->{str} = $gen_types{ $parm->{type} }{str};
            $parm->{sqltype} = to_SQL_type($parm->{type},$parm->{size});
            $parm->{cstype} = $gen_types->{$parm->{type}}{cs};
            $parm->{csdbtype} = $gen_types->{$parm->{type}}{csdb};
            push @{$attr->{param}},$parm;
        }
        push @{$obj->{squery}},$attr;
    }




    $obj->{value} = queryXML($object,'object/value/$');
    $obj->{display} = queryXML($object,'object/display/$');

    $obj->{value} = $obj->{attrib_names}{id} if !$obj->{value};
    $obj->{display} = $obj->{attrib_names}{name} if !$obj->{display};

    return $obj;
}


#################################################################################################
## convert my type into an SQL type
#################################################################################################
sub to_SQL_type {
    my ($type,$size) = @_;
    my $v = $gen_types->{$type}{sql};
    $v .= "($size)" if $size > 0;
    return $v;
}















######################################################################
## returns contents of file
######################################################################
sub loadFile {
   my ($name) = @_;
   open my $in,$name or die qq{$!: "$name"};

   read $in,my $cont,-s $name or die qq{$!: "$name"};
   close $in;
   return $cont;
}

######################################################################
## XML Parser & Query Code
######################################################################

# xml parser; returns root node
sub parseXML {
    $_ = shift;
    my @strings;
    s{<!\[CDATA\[(.*?)\]\]>}{push @strings,$1;'<#'.$#strings.'/>'}xsge;
    s|<!--.*?-->||gs;
    push my @stack, { type=>root, value=>'~', children=>[]};
    while(m|(.*?)<(.+?)>|gs){
        for(my $t = $1){
            s/^\s+|\s+$//gs; s/\s+/ /gs; s|\&lt;|<|gs; s|\&gt;|>|gs; 
            s|\&amp;|\&|gs; s|\&apos;|\'|gs; s|\&quot;|\"|gs;
            push @{ $stack[$#stack]->{children} },{type=>text,value=>$_} if length;
        }
        my $t = $2;
        if($t !~ m|/|){ # opening tag
            my ($tag,$attr) = split /\s+/,$t,2;
            push @stack, { type=>tag, value=>$tag, attrib=>{}, children=>[] } ;
            $stack[$#stack]->{attrib}{$1} = $2 while($attr =~ m|\s*(.*?)="(.*?)"|g);
        }elsif($t =~ m|^/|){ # closing tag
            my $s = pop @stack;
            push @{ $stack[$#stack]->{children} },$s if $t eq "/$s->{value}";
        }elsif($t =~ m|#(\d+)|){    # CDATA section
            push @{ $stack[$#stack]->{children} },{type=>text,value=>$strings[$1]} if $strings[$1];
        }elsif($t =~ m|/$|){
            $t =~ s|/$||;
            my ($tag,$attr) = split /\s+/,$t,2;
            my $o = {type=>tag,value=>$tag, attrib=>{}, children=>[]};
            $o->{attrib}{$1} = $2 while($attr =~ m|\s*(.*?)="(.*?)"|g);
            push @{ $stack[$#stack]->{children} },$o;
        }
    }
    return $stack[0];
}

######################################################################
## given root node, displays the XML
######################################################################
sub dispXML {
    my $root = shift;
    my @stack;
    for( reverse @{ $root->{children} } ){
        push @stack,[$_,0];
    }
    while($#stack >= 0){
        my ($tag,$tab) = @{pop @stack};
        if($tag->{type} eq tag){
            print ' 'x $tab,"<",$tag->{value};
            for(sort keys %{$tag->{attrib}}){
                print ' ',$_,'="',$tag->{attrib}{$_},'"';
            }
            print ">\n";
            push @stack,[ { type=>ctag, value=>$tag->{value} },$tab];
            for( reverse @{ $tag->{children} } ){
                push @stack,[$_,$tab+3];
            }
        }elsif($tag->{type} eq ctag){
            print ' 'x $tab,'</',$tag->{value},">\n";
        }elsif($tag->{type} eq text){
            print ' 'x $tab,$tag->{value},"\n";
        }
    }
}


######################################################################
# ---queries---
#  schema/object/%     (retrieves all attributes)
#  schema/object/%name (retrieves name attribute for each object)
#  schema/object/@      (retrieves all children)
#  schema/@object       (retrieves all object children)
#  schema/object/$     (retrieves any text in each object)
#  /schema/@object      (ensures that we start search from root)
#  schema/*/@     (* indicates any tag)
# query XML for various tags/attributes
######################################################################
sub queryXML {
    my ($root,$query) = @_;
    my @query = split '/',$query;

    my @result;

    # prepare query
    $query[0] = '~' if !$query[0];
    my $result = pop @query;       # what are we looking for?
    $query = join('#',@query).'#';
    $query =~ s/\*/[^#]+?/;

#    print "query: $query\n";

    push my @stack,[$root,0];

    while($#stack >= 0){
        my ($tag,$depth) = @{pop @stack};

        if($tag->{type} ne text){
            splice @current, $depth;
            push @current, $tag->{value};
            $current = join('#',@current).'#';

#            print "compare: $current =~ m/$query/\n";

            if($current =~ m/$query/){

#                print "($result): $current =~ m/$query/\n";

                if($result =~ m/\@(.*)/){
                    my $n = $1;
                    if($n){
                        for( @{ $tag->{children} } ){
                            if($_->{type} eq tag && $_->{value} eq $n){
                                push @result,$_;
                            }
                        }
                    }else{
                        for( @{ $tag->{children} } ){
                            if($_->{type} eq tag){
                                push @result,$_;
                            }
                        }
                    }
                }elsif($result =~ m/\%(.*)/){
                    my $n = $1;
                    if($n){
                        push @result,$tag->{attrib}{$n} if $tag->{attrib}{$n};
                    }else{
                        for(sort keys %{$tag->{attrib}}){
                            push @result,[$_,$tag->{attrib}{$_}];
                        }
                    }
                }elsif($result =~ m/\$(.*)/){
                    for( @{ $tag->{children} } ){
                        push @result,$_->{value} if $_->{type} eq text;
                    }
                }

            }else{
                for( @{ $tag->{children} } ){
                    push @stack,[$_,$depth+1];
                }
            }
        }
    }
    return wantarray ? @result : $result[0];
}

